/* -*- Mode: C++ -*-
 *Header:
 *File: object.C
 *Author: Noda Itsuki
 *Date: 1996/01/28
 *EndHeader:
 */

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Form.h>

#include <math.h>
#include <stdio.h>
#include <iostream.h>

#include "param.h"
#include "fallback.h"
#include "types.h"
#include "utility.h"
#include "object.h"
#include "netif.h"
#include "field.h"
#include "ball.xbm"

extern Stadium *Std ;

#ifdef X11R5
static Arg Args[16] ;
static int Argn ;
#endif

/*
 *===================================================================
 *Part: Ball
 *===================================================================
 */
void Ball::assign(void)
{
	int status ;
	unsigned int ball_xbm_width, ball_xbm_height ;

	Pixmap pix, tmp ;
	GC	gc ;
	XGCValues gcv ;
	XColor def, closest ;
	disp = XtDisplay(Std->field.margin) ;
	Window win = XtWindow(Std->field.margin) ;
	Screen *scr = XtScreen(Std->field.margin) ;

	XAllocNamedColor(disp, DefaultColormapOfScreen(scr),
										DEFAULT_FIELD_COLOR, &def, &closest) ;

	if ((status = XReadBitmapFile(disp, win,
									Std->ball_file,
									&ball_xbm_width, &ball_xbm_height, &tmp,
									NULL, NULL)) == BitmapSuccess) {
		pix = XCreatePixmap(disp, win, ball_xbm_width, ball_xbm_height,
												DefaultDepthOfScreen(scr)) ;
		gcv.function = GXcopy ;
		gcv.foreground = WhitePixel(disp, 0) ;
		gcv.background = closest.pixel ;

		gc = XCreateGC(disp, pix, (GCFunction|GCForeground|GCBackground), &gcv);

		XCopyPlane(disp, tmp, pix, gc,
					0, 0, ball_xbm_width, ball_xbm_height, 0, 0, 1) ;
		XFreeGC(disp, gc) ;
		XFreePixmap(disp, tmp) ;
	}
	else {
		switch(status){
		case BitmapOpenFailed:
			fprintf(stderr, "xbitmap: could not open ball bitmap file.\n") ;
			break ;
		case BitmapFileInvalid:
			fprintf(stderr, "xbitmap: ball bitmap file invalid.\n") ;
			break ;
		case BitmapNoMemory:
			fprintf(stderr, "xbitmap: insufficient server memory to create ball bitmap.\n") ;
			break ;
		}
		fprintf(stderr, "         use default bitmap file.\n") ;
		pix = XCreatePixmapFromBitmapData(disp,DefaultRootWindow(disp),
								ball_bits,ball_width,ball_height,
								WhitePixel(disp, 0), closest.pixel,
								DefaultDepthOfScreen(scr)) ;
		ball_xbm_width = ball_width ;
		ball_xbm_height = ball_height ;
	}

	ball = XCreateSimpleWindow(disp, win,
							(Position)ball_xbm_width, (Position)ball_xbm_height,
							ball_xbm_width, ball_xbm_height, 0, 0, 0) ;
	XSetWindowBackgroundPixmap(disp, ball, pix) ;
	XMapWindow(disp, ball) ;
	XFreePixmap(disp, pix) ;

	length_magnify = Std->length_magnify ;
	pos_x = -1 ;
	pos_y = -1 ;
}

void Ball::set(short x, short y)
{
	if (pos_x != x || pos_y != y) {
		XMoveWindow(disp, ball, 
				max(InFieldPosX((x/SHOWINFO_SCALE)) - ball_width/2,0),
				max(InFieldPosY((y/SHOWINFO_SCALE)) - ball_height/2,0)
				) ;
		pos_x = x ;
		pos_y = y ;
		XFlush(disp) ;
	}
}

/*
 *===================================================================
 *Part: Player
 *===================================================================
 */
void Player::assign(Side side, Name team, int unum, short kind)
{
	XColor def, closest ;
	Window win = XtWindow(Std->field.margin) ;
	disp = XtDisplay(Std->field.margin) ;

	length_magnify = Std->length_magnify ;
	pfont_x = Std->pfont_x ;
	pfont_y = Std->pfont_y ;
	size = max(((Dimension)Std->psize),1) ;
	sprintf(name, PLAYER_NAME_FORMAT, team, unum) ;
	sprintf(unumstr, "%d", unum) ;
	ang = 1000 ;
	pos_x = -1 ;
	pos_y = -1 ;
	prev_kind = DISABLE ;

	XColor c0, c1 ;
	XGCValues gcv ;

	pix = XCreatePixmap(disp, win, size, size, DefaultDepth(disp, 0)) ;

	gcv.function = GXcopy ;

	XAllocNamedColor(disp, DefaultColormap(disp, 0),
					(((side)== LEFT)? Std->color_l:Std->color_r), &c1, &c0) ;
	gcv.foreground = c1.pixel ;
	fgh = XCreateGC(disp, pix, (GCFunction|GCForeground), &gcv) ;

	XAllocNamedColor(disp, DefaultColormap(disp, 0),
		(((side)== LEFT)? Std->goalie_color_l:Std->goalie_color_r), &c1, &c0) ;
	gcv.foreground = c1.pixel ;
	fgh_g = XCreateGC(disp, pix, (GCFunction|GCForeground), &gcv) ;

	XAllocNamedColor(disp, DefaultColormap(disp, 0),
				DEFAULT_PLAYER_COLOR_T, &c1, &c0) ;
	gcv.foreground = c1.pixel ;
	fgt = XCreateGC(disp, pix, (GCFunction|GCForeground), &gcv) ;

	XAllocNamedColor(disp, DefaultColormap(disp, 0),
				DEFAULT_FIELD_COLOR, &c1, &c0) ;
	gcv.foreground = c1.pixel ;
	bg = XCreateGC(disp, pix, (GCFunction|GCForeground), &gcv) ;

	XAllocNamedColor(disp, DefaultColormap(disp, 0),
				DEFAULT_PLAYER_COLOR_S, &c1, &c0) ;

	XFontStruct *font ;
	if ((font = XLoadQueryFont(disp, Std->pfont)) == NULL) {
		printf("Can not find Player Widget Font!\nUse Default Font.\n") ;
		strcpy(Std->pfont, DEFAULT_PLAYER_FONT) ;
	}
	else
		XFreeFont(disp, font) ;

	gcv.font = XLoadFont(disp, Std->pfont) ;
	gcv.foreground = c1.pixel ;

	fgs = XCreateGC(disp, pix, (GCFunction|GCForeground|GCFont), &gcv) ;


	player = XCreateSimpleWindow(disp, win,
								(Position)size, (Position)size, 
								size+1, size+1, 0, 0, 0) ;

	XMapWindow(disp, player) ;
}

void Player::set(short x, short y, short angle, short kind)
{
	if (ang != angle || prev_kind != kind) {
		XFillRectangle( disp, pix, bg, 0,0, size, size) ;
		if (!(kind & STAND))
			XFillArc(disp, pix, fgt, 0, 0, size, size, 0, 360*64) ;
		else if (kind & GOALIE){
			XFillArc(disp, pix, fgh_g, 0, 0, size, size, 0, 360*64) ;
			if (!(kind & CATCH))
				XFillArc(disp, pix, fgt,
									0, 0, size, size, (-angle+90)*64, 180*64) ;
			XDrawArc(disp, pix, fgh_g, 0, 0, size, size, 0,360*64) ;
			if (kind & KICK) {
				XDrawArc(disp, pix, fgh_g, 2, 2, size-4, size-4, 0,360*64) ;
				XDrawArc(disp, pix, fgh_g, 3, 3, size-6, size-6, 0,360*64) ;
				XDrawArc(disp, pix, fgh_g, 4, 4, size-8, size-8, 0,360*64) ;
			}
		}
		else {
			XFillArc(disp, pix, fgh, 0, 0, size, size, 0, 360*64) ;
			XFillArc(disp, pix, fgt, 0, 0, size, size, (-angle+90)*64, 180*64) ;
			XDrawArc(disp, pix, fgh, 0, 0, size, size, 0,360*64) ;
			if (kind & KICK) {
				XDrawArc(disp, pix, fgh, 2, 2, size-4, size-4, 0,360*64) ;
				XDrawArc(disp, pix, fgh, 3, 3, size-6, size-6, 0,360*64) ;
				XDrawArc(disp, pix, fgh, 4, 4, size-8, size-8, 0,360*64) ;
			}
		}
		XDrawString(disp, pix, fgs, pfont_x, pfont_y, unumstr,strlen(unumstr)) ;
		XSetWindowBackgroundPixmap(disp, player, pix) ;
		XClearWindow(disp, player);

		ang = angle ;
		prev_kind = kind ;
	}

	if (pos_x != x || pos_y != y) {
		XMoveWindow(disp, player, 
					max(InFieldPosX((x/SHOWINFO_SCALE)) - size/2,0),
					max(InFieldPosY((y/SHOWINFO_SCALE)) - size/2,0)
					) ;
		pos_x = x ;
		pos_y = y ;
		XFlush(disp) ;
	}
}

void Player::destroy(void)
{
	XDestroyWindow(disp, player) ;
	XFreePixmap(disp, pix) ;
	XFreeGC(disp, fgh) ;
	XFreeGC(disp, fgt) ;
	XFreeGC(disp, fgs) ;
	XFreeGC(disp, bg) ;
	XFlush(disp) ;
	player = FALSE ;
}
